from typing import Any

from googleapiclient.errors import HttpError

from gaclw.client.base import GoogleApiClient


class GoogleSpreadsheetsApiClient(GoogleApiClient):
    def __init__(self, log_level: int = 2):
        scopes = [
            "https://www.googleapis.com/auth/spreadsheets",
        ]
        super().__init__("sheets", "v4", scopes, log_level)
        self._spreadsheets = self.service.spreadsheets()  # type: ignore

    @property
    def spreadsheets(self):
        return self._spreadsheets

    def append_values(
        self, spreadsheet_id: str, range_name: str, values: list[list[Any]]
    ) -> None:
        try:
            result = (
                self.spreadsheets.values()
                .append(
                    spreadsheetId=spreadsheet_id,
                    range=range_name,
                    body={"values": values},
                    valueInputOption="USER_ENTERED",
                    insertDataOption="INSERT_ROWS",
                )
                .execute()
            )
            self.logger.info(
                f"Updated Range: {result.get('updates').get('updatedRange')}"
            )
            self.logger.info(f"Updated Rows: {result.get('updates').get('updatedRows')}")
            self.logger.info(
                f"Updated Columns: {result.get('updates').get('updatedColumns')}"
            )
            self.logger.info(
                f"Updated Cells: {result.get('updates').get('updatedCells')}"
            )
        except HttpError as err:
            self.logger.error(err)
            return None

    def get_values(self, spreadsheet_id: str, range_name: str) -> list[list[Any]] | None:
        try:
            result = (
                self.spreadsheets.values()
                .get(
                    spreadsheetId=spreadsheet_id,
                    range=range_name,
                )
                .execute()
            )
            self.logger.info(f"Major dimension: {result.get('majorDimension')}")
            self.logger.info(f"Range: {result.get('range')}")
            return result.get("values")
        except HttpError as err:
            self.logger.error(err)
            return None



"""client module"""

from gaclw.client.drive import GoogleDriveApiClient
from gaclw.client.spreadsheet import GoogleSpreadsheetsApiClient


__all__ = ["GoogleDriveApiClient", "GoogleSpreadsheetsApiClient"]



"""Base class for clients"""

import logging
import sys
from pathlib import Path

from google.auth.external_account_authorized_user import (
    Credentials as ExternalAccountCredentials,
)
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import Resource, build

from gaclw.core.config import settings


class GoogleApiClient(object):
    """Base class for clients"""

    # https://developers.google.com/drive/api/guides/api-specific-auth
    # https://developers.google.com/sheets/api/scopes?hl=ja
    TOKEN_FILE_PATH = Path(settings.SECRET_FOLDER, "token.json")
    CREDENTIALS_FILE_PATH = Path(settings.SECRET_FOLDER, "credentials.json")

    def __init__(
        self,
        service_name: str,
        version: str,
        scopes: list[str],
        log_level: int = 2,
    ):
        self._logger = self.get_logger(log_level)
        self._credentials = self.authenticate(scopes)
        self._service = build(service_name, version, credentials=self._credentials)

    @property
    def logger(self) -> logging.Logger:
        """_summary_

        Returns:
            logging.Logger: _description_
        """
        return self._logger

    @property
    def service(self) -> Resource:
        """_summary_

        Returns:
            Resource: _description_
        """
        return self._service

    @classmethod
    def get_logger(cls, level: int = 2) -> logging.Logger:
        """_summary_

        Args:
            level (int, optional): _description_. Defaults to 2.

        Returns:
            logging.Logger: _description_
        """
        logger = logging.getLogger(__name__)
        logger.setLevel(level * 10)
        formatter = logging.Formatter(
            "%(asctime)s [%(levelname)s] %(message)s",
            datefmt=r"%Y-%m-%dT%H:%M:%S%z",
        )
        # ログを標準出力する
        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger

    @classmethod
    def authenticate(cls, scopes: list[str]) -> Credentials | ExternalAccountCredentials:
        """_summary_

        Args:
            scopes (list[str]): _description_

        Returns:
            Credentials | ExternalAccountCredentials: _description_
        """
        credentials = None
        if cls.TOKEN_FILE_PATH.exists():
            credentials = Credentials.from_authorized_user_file(
                cls.TOKEN_FILE_PATH, scopes
            )

        if not credentials or not credentials.valid:
            if credentials and credentials.expired and credentials.refresh_token:
                credentials.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    cls.CREDENTIALS_FILE_PATH, scopes
                )
                credentials = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open(cls.TOKEN_FILE_PATH, "w", encoding="utf-8") as token:
                token.write(credentials.to_json())
        return credentials



"""define GoogleDriveApiClient class"""

import pathlib

from googleapiclient.errors import HttpError
from googleapiclient.http import MediaFileUpload

from gaclw.client.base import GoogleApiClient


class GoogleDriveApiClient(GoogleApiClient):
    """Google Drive API client"""

    def __init__(self, log_level: int = 2):
        scopes = ["https://www.googleapis.com/auth/drive"]
        super().__init__("drive", "v3", scopes, log_level)
        self._files = self.service.files()  # type: ignore

    @property
    def files(self):
        return self._files

    def find_folder(
        self, folder_name: str, parent_folder_id: str | None = None
    ) -> str | None:
        """Find folder by name and parent folder ID in Google Drive."""
        query = (
            f"name='{folder_name}'"
            + " and trashed=false"
            + " and mimeType='application/vnd.google-apps.folder'"
        )
        if parent_folder_id:
            query += f" and '{parent_folder_id}' in parents"
        results = self.files.list(
            q=query,
            fields="files(id, name)",
            supportsTeamDrives=True,
            includeItemsFromAllDrives=True,
        ).execute()

        try:
            items: list[dict[str, str]] = results.get("files", [])
            self.logger.warning(
                "Found multiple folders with the same name: %s", folder_name
            )
            for item in results.get("files", []):
                self.logger.debug("Found folder: %s (ID: %s)", item["name"], item["id"])
            return next(i["id"] for i in items)
        except StopIteration:
            self.logger.warning("No folder found with the name: %s", folder_name)
            return None

    def _find_file(
        self, file_name: str, parent_folder_id: str | None = None
    ) -> str | None:
        query = (
            f"name='{file_name}'"
            + " and trashed=false"
            + " and mimeType!='application/vnd.google-apps.folder'"
        )
        if parent_folder_id:
            query += f" and '{parent_folder_id}' in parents"
        results = self.files.list(
            q=query,
            fields="files(id, name)",
            supportsTeamDrives=True,
            includeItemsFromAllDrives=True,
        ).execute()

        try:
            items: list[dict[str, str]] = results.get("files", [])
            self.logger.warning("Found multiple files with the same name: %s", file_name)
            for item in results.get("files", []):
                self.logger.debug("Found file: %s (ID: %s)", item["name"], item["id"])
            return next(i["id"] for i in items)
        except StopIteration:
            self.logger.warning("No file found with the name: %s", file_name)
            return None

    def upload_file(
        self, file_path: pathlib.Path, parent_folder_id: str | None
    ) -> str | None:
        """Upload file to Google Drive."""
        metadata = {
            "name": file_path.name,
            "parents": [] if parent_folder_id is None else [parent_folder_id],
        }
        media = MediaFileUpload(file_path, resumable=True)

        try:
            existing_id = self._find_file(file_path.name, parent_folder_id)
            if existing_id:
                _ = metadata.pop("parents")
                uploaded_file: dict[str, str] = self.files.update(
                    fileId=existing_id,
                    body=metadata,
                    media_body=media,
                    removeParents=f"{parent_folder_id}",
                    addParents=f"{parent_folder_id}",
                    supportsTeamDrives=True,
                ).execute()
                self.logger.info(
                    "Updated file: %s with ID: %s",
                    file_path.name,
                    uploaded_file.get("id"),
                )
            else:
                uploaded_file = self.files.create(
                    body=metadata,
                    media_body=media,
                    fields="id",
                    supportsTeamDrives=True,
                ).execute()
                self.logger.info(
                    "Uploaded file: %s with ID: %s",
                    file_path.name,
                    uploaded_file.get("id"),
                )
            return uploaded_file.get("id")
        except HttpError as error:
            self.logger.error("An error occurred: %s", error)
            return None

    def create_folder(
        self, folder_name: str, parent_folder_id: str | None = None
    ) -> str | None:
        """Create folder in Google Drive."""
        metadata = {
            "name": folder_name,
            "mimeType": "application/vnd.google-apps.folder",
            "parents": [] if parent_folder_id is None else [parent_folder_id],
        }
        try:
            existing_id = self.find_folder(folder_name, parent_folder_id)
            if existing_id:
                _ = metadata.pop("parents")
                created_folder: dict[str, str] = self.files.update(
                    fileId=existing_id,
                    body=metadata,
                    removeParents=f"{parent_folder_id}",
                    addParents=f"{parent_folder_id}",
                    supportsTeamDrives=True,
                ).execute()
                self.logger.info(
                    "Updated folder: %s with ID: %s",
                    folder_name,
                    created_folder.get("id"),
                )
            else:
                created_folder = self.files.create(
                    body=metadata, fields="id", supportsTeamDrives=True
                ).execute()
                self.logger.info(
                    "Created folder: %s with ID: %s",
                    folder_name,
                    created_folder.get("id"),
                )
        except HttpError as error:
            self.logger.error("An error occurred: %s", error)
            return None
        else:
            return created_folder.get("id", "NA")



"""Configuration"""

from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Configuration"""

    model_config = SettingsConfigDict(
        # Use top level .env file
        env_file=".env",
        env_ignore_empty=True,
        extra="ignore",
    )

    SECRET_FOLDER: str


settings = Settings()  # type: ignore
